/*
 * Copyright (c) 2016-2019 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Tue Aug 23 2016
 */
#include "acmrsim.h"

#include <sys/time.h>
#include <math.h>

#include "mrsim_c_types.h"
#include "codels.h"


/* --- mrsim_update_state -------------------------------------------------- */

void
mrsim_update_state(const mrsim_s *mrsim, or_pose_estimator_state *state)
{
  struct mrsimstate_body_s *body = mrsim_body(mrsim);

  /* ground reaction */
  if (body->pz < 0.) {
    body->pz = 0.;
    body->qw = sqrt(1 - body->qz*body->qz); body->qx = body->qy = 0.;
    body->vx = body->vy = body->vz = 0.;
    body->wx = body->wy = body->wz = 0.;
    body->ax = body->ay = body->az = 0.;
    body->dwx = body->dwy = body->dwz = 0.;
  }

  state->pos._value.x = body->px;
  state->pos._value.y = body->py;
  state->pos._value.z = body->pz;

  state->att._value.qw = body->qw;
  state->att._value.qx = body->qx;
  state->att._value.qy = body->qy;
  state->att._value.qz = body->qz;
  {
    double qvar =
      mrsim_params(mrsim)->state.qstddev * mrsim_params(mrsim)->state.qstddev;
    double qw = state->att._value.qw;
    double qx = state->att._value.qx;
    double qy = state->att._value.qy;
    double qz = state->att._value.qz;

    state->att_cov._value.cov[0] = qvar * (1 - qw*qw);
    state->att_cov._value.cov[1] = qvar * -qw*qx;
    state->att_cov._value.cov[2] = qvar * (1 - qx*qx);
    state->att_cov._value.cov[3] = qvar * qw*qy;
    state->att_cov._value.cov[4] = qvar * -qx*qy;
    state->att_cov._value.cov[5] = qvar * (1 - qy*qy);
    state->att_cov._value.cov[6] = qvar * -qw*qz;
    state->att_cov._value.cov[7] = qvar * -qx*qz;
    state->att_cov._value.cov[8] = qvar * -qy*qz;
    state->att_cov._value.cov[9] = qvar * (1 - qz*qz);
  }

  state->vel._value.vx = body->vx;
  state->vel._value.vy = body->vy;
  state->vel._value.vz = body->vz;

  state->avel._value.wx = body->wx;
  state->avel._value.wy = body->wy;
  state->avel._value.wz = body->wz;

  state->acc._value.ax = body->ax;
  state->acc._value.ay = body->ay;
  state->acc._value.az = body->az;
}


/* --- mrsim_addnoise ------------------------------------------------------ */

void
mrsim_addnoise(const mrsim_s *mrsim, or_pose_estimator_state *state)
{
  struct mrsim_param_s *p = mrsim_params(mrsim);

  if (state->pos._present) {
    if (p->state.pstddev > 0.) {
      state->pos._value.x += mrsim_gnoise(p->state.pstddev);
      state->pos._value.y += mrsim_gnoise(p->state.pstddev);
      state->pos._value.z += mrsim_gnoise(p->state.pstddev);
    }

    if (p->state.qstddev > 0.) {
      double nw;
      double nx = mrsim_gnoise(p->state.qstddev);
      double ny = mrsim_gnoise(p->state.qstddev);
      double nz = mrsim_gnoise(p->state.qstddev);
      double n = sqrt(nx*nx + ny*ny + nz*nz);
      if (n > 0.) {
        double qw = state->att._value.qw;
        double qx = state->att._value.qx;
        double qy = state->att._value.qy;
        double qz = state->att._value.qz;

        nw = cos(n/2);
        nx *= sin(n/2)/n;
        ny *= sin(n/2)/n;
        nz *= sin(n/2)/n;

        state->att._value.qw = qw*nw - qx*nx - qy*ny - qz*nz;
        state->att._value.qx = qw*nx + qx*nw + qy*nz - qz*ny;
        state->att._value.qy = qw*ny - qx*nz + qy*nw + qz*nx;
        state->att._value.qz = qw*nz + qx*ny - qy*nx + qz*nw;
      }
    }
  }

  if (state->avel._present && p->state.wstddev > 0.) {
    state->avel._value.wx += mrsim_gnoise(p->state.wstddev);
    state->avel._value.wy += mrsim_gnoise(p->state.wstddev);
    state->avel._value.wz += mrsim_gnoise(p->state.wstddev);
  }

  if (state->acc._present && p->state.astddev > 0.) {
    state->acc._value.ax += mrsim_gnoise(p->state.astddev);
    state->acc._value.ay += mrsim_gnoise(p->state.astddev);
    state->acc._value.az += mrsim_gnoise(p->state.astddev);
  }
}
